// DeCross.dll v0.0.02 Copyright(C) 2003, 2004 minamina
// Avisynth Plugin - Cross Color Reduction (YUY2 and YV12 Only)

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "DeCross.h"
#pragma warning (push, 1)
#pragma warning (disable : 4799)
#include "ivec_wrapper.h"
#pragma warning (pop)

extern "C" __declspec(dllexport)
const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* env)
{
	DeCross::PluginInit(env);
	return 0;
}

DeCross::DeCross(PClip _child, const int nYThreshold, const int nNoiseThreshold,
				 const int nMargin, const bool bDebug, IScriptEnvironment* env)
	: GenericVideoFilter(_child)
{
	if(!(vi.IsYUY2() || vi.IsYV12()))
	{
		env->ThrowError("%s : YUY2, YV12 only", GetName());
	}

	m_nYThreshold = min(max(0, nYThreshold), 255);
	m_nNoiseThreshold = min(max(0, nNoiseThreshold), 255);
	m_nMargin = min(max(0, nMargin), 4);
	m_bDebug = bDebug;
}

PVideoFrame __stdcall DeCross::GetFrame(int n, IScriptEnvironment *env)
{
	if (n == 0 || n >= vi.num_frames - 1)
	{
		return child->GetFrame(n, env);
	}

	PVideoFrame destFrame;

	if (vi.IsYUY2())
	{
		destFrame = DeCross_YUY2(n, env);
	}
	else
	{
		destFrame = DeCross_YV12(n, env);
	}
	return destFrame;
}

PVideoFrame DeCross::DeCross_YUY2(int n, IScriptEnvironment *env)
{
	PVideoFrame srcFrame = child->GetFrame(n, env);
	PVideoFrame destFrame = srcFrame;
	env->MakeWritable(&destFrame);

	const int nHeight = srcFrame->GetHeight();
	const int nRowSize = srcFrame->GetRowSize();
	const int nSrcPitch = srcFrame->GetPitch();
	const int nSrcPitch2 = nSrcPitch * 2;
	const int nDestPitch = destFrame->GetPitch();

	const BYTE* pSrc = srcFrame->GetReadPtr() + nSrcPitch2;
	const BYTE* pSrcP = child->GetFrame(n - 1, env)->GetReadPtr() + nSrcPitch2;
	const BYTE* pSrcF = child->GetFrame(n + 1, env)->GetReadPtr() + nSrcPitch2;

	const BYTE* pSrcTT = pSrc - nSrcPitch2;
	const BYTE* pSrcT = pSrc - nSrcPitch;
	const BYTE* pSrcB = pSrc + nSrcPitch;
	const BYTE* pSrcBB = pSrc + nSrcPitch2;
	const BYTE* pSrcPTT = pSrcP - nSrcPitch2;
	const BYTE* pSrcPT = pSrcP - nSrcPitch;
	const BYTE* pSrcPB = pSrcP + nSrcPitch;
	const BYTE* pSrcPBB = pSrcP + nSrcPitch2;
	const BYTE* pSrcFTT = pSrcF - nSrcPitch2;
	const BYTE* pSrcFT = pSrcF - nSrcPitch;
	const BYTE* pSrcFB = pSrcF + nSrcPitch;
	const BYTE* pSrcFBB = pSrcF + nSrcPitch2;

	BYTE* pDest = destFrame->GetWritePtr() + nDestPitch * 2;

	BYTE* pEdgeBuffer = new BYTE[nRowSize];

	int nX, nY;
	for (nY = nHeight - 2; nY > 2; nY--)
	{
		::ZeroMemory(pEdgeBuffer, nRowSize);

		EdgeCheck_YUY2(pSrc, pEdgeBuffer, nRowSize);

		if (m_bDebug)
		{
			for (nX = 16; nX < nRowSize - 16; nX += 4)
			{
				if (pEdgeBuffer[nX] != 0)
				{
					pDest[nX + 1] = 128;
					pDest[nX + 3] = 255;
				}
			}
		}
		else
		{
			for (nX = 16; nX < nRowSize - 16; nX += 16)
			{
				if ((*(DWORD*)&pEdgeBuffer[nX] != 0 || *(DWORD*)&pEdgeBuffer[nX + 4] != 0 ||
					*(DWORD*)&pEdgeBuffer[nX + 8] != 0 || *(DWORD*)&pEdgeBuffer[nX + 12] != 0))
				{
					int nMiniDiff = m_nNoiseThreshold;
					const BYTE* pMiniSrc = pSrc + nX;

					Diff_YUY2(pSrcPTT + nX, pSrc + nX, -6, nMiniDiff, pMiniSrc);
					Diff_YUY2(pSrcPTT + nX, pSrc + nX, -2, nMiniDiff, pMiniSrc);
					Diff_YUY2(pSrcPT  + nX, pSrc + nX, -4, nMiniDiff, pMiniSrc);
					Diff_YUY2(pSrcP   + nX, pSrc + nX, -4, nMiniDiff, pMiniSrc);
					Diff_YUY2(pSrcPB  + nX, pSrc + nX, -6, nMiniDiff, pMiniSrc);
					Diff_YUY2(pSrcPB  + nX, pSrc + nX, -2, nMiniDiff, pMiniSrc);
					Diff_YUY2(pSrcPBB + nX, pSrc + nX, -6, nMiniDiff, pMiniSrc);
					Diff_YUY2(pSrcPBB + nX, pSrc + nX, -2, nMiniDiff, pMiniSrc);
					Diff_YUY2(pSrcTT  + nX, pSrc + nX, -4, nMiniDiff, pMiniSrc);
					Diff_YUY2(pSrcT   + nX, pSrc + nX, -6, nMiniDiff, pMiniSrc);
					Diff_YUY2(pSrcT   + nX, pSrc + nX, -2, nMiniDiff, pMiniSrc);
					Diff_YUY2(pSrcB   + nX, pSrc + nX, -4, nMiniDiff, pMiniSrc);
					Diff_YUY2(pSrcBB  + nX, pSrc + nX, -4, nMiniDiff, pMiniSrc);
					Diff_YUY2(pSrcFTT + nX, pSrc + nX, -6, nMiniDiff, pMiniSrc);
					Diff_YUY2(pSrcFTT + nX, pSrc + nX, -2, nMiniDiff, pMiniSrc);
					Diff_YUY2(pSrcFT  + nX, pSrc + nX, -4, nMiniDiff, pMiniSrc);
					Diff_YUY2(pSrcF   + nX, pSrc + nX, -4, nMiniDiff, pMiniSrc);
					Diff_YUY2(pSrcFB  + nX, pSrc + nX, -6, nMiniDiff, pMiniSrc);
					Diff_YUY2(pSrcFB  + nX, pSrc + nX, -2, nMiniDiff, pMiniSrc);
					Diff_YUY2(pSrcFBB + nX, pSrc + nX, -6, nMiniDiff, pMiniSrc);
					Diff_YUY2(pSrcFBB + nX, pSrc + nX, -2, nMiniDiff, pMiniSrc);

					Diff_YUY2(pSrcPT + nX, pSrc + nX, 0, nMiniDiff, pMiniSrc);
					Diff_YUY2(pSrcP  + nX, pSrc + nX, 0, nMiniDiff, pMiniSrc);
					Diff_YUY2(pSrcTT + nX, pSrc + nX, 0, nMiniDiff, pMiniSrc);
					Diff_YUY2(pSrcT  + nX, pSrc + nX, 0, nMiniDiff, pMiniSrc);
					Diff_YUY2(pSrcB  + nX, pSrc + nX, 0, nMiniDiff, pMiniSrc);
					Diff_YUY2(pSrcBB + nX, pSrc + nX, 0, nMiniDiff, pMiniSrc);
					Diff_YUY2(pSrcFT + nX, pSrc + nX, 0, nMiniDiff, pMiniSrc);
					Diff_YUY2(pSrcF  + nX, pSrc + nX, 0, nMiniDiff, pMiniSrc);

					Diff_YUY2(pSrcPTT + nX, pSrc + nX, 6, nMiniDiff, pMiniSrc);
					Diff_YUY2(pSrcPTT + nX, pSrc + nX, 2, nMiniDiff, pMiniSrc);
					Diff_YUY2(pSrcPT  + nX, pSrc + nX, 4, nMiniDiff, pMiniSrc);
					Diff_YUY2(pSrcP   + nX, pSrc + nX, 4, nMiniDiff, pMiniSrc);
					Diff_YUY2(pSrcPB  + nX, pSrc + nX, 6, nMiniDiff, pMiniSrc);
					Diff_YUY2(pSrcPB  + nX, pSrc + nX, 2, nMiniDiff, pMiniSrc);
					Diff_YUY2(pSrcPBB + nX, pSrc + nX, 6, nMiniDiff, pMiniSrc);
					Diff_YUY2(pSrcPBB + nX, pSrc + nX, 2, nMiniDiff, pMiniSrc);
					Diff_YUY2(pSrcTT  + nX, pSrc + nX, 4, nMiniDiff, pMiniSrc);
					Diff_YUY2(pSrcT   + nX, pSrc + nX, 6, nMiniDiff, pMiniSrc);
					Diff_YUY2(pSrcT   + nX, pSrc + nX, 2, nMiniDiff, pMiniSrc);
					Diff_YUY2(pSrcB   + nX, pSrc + nX, 4, nMiniDiff, pMiniSrc);
					Diff_YUY2(pSrcBB  + nX, pSrc + nX, 4, nMiniDiff, pMiniSrc);
					Diff_YUY2(pSrcFTT + nX, pSrc + nX, 6, nMiniDiff, pMiniSrc);
					Diff_YUY2(pSrcFTT + nX, pSrc + nX, 2, nMiniDiff, pMiniSrc);
					Diff_YUY2(pSrcFT  + nX, pSrc + nX, 4, nMiniDiff, pMiniSrc);
					Diff_YUY2(pSrcF   + nX, pSrc + nX, 4, nMiniDiff, pMiniSrc);
					Diff_YUY2(pSrcFB  + nX, pSrc + nX, 6, nMiniDiff, pMiniSrc);
					Diff_YUY2(pSrcFB  + nX, pSrc + nX, 2, nMiniDiff, pMiniSrc);
					Diff_YUY2(pSrcFBB + nX, pSrc + nX, 6, nMiniDiff, pMiniSrc);
					Diff_YUY2(pSrcFBB + nX, pSrc + nX, 2, nMiniDiff, pMiniSrc);

					I8vec8 mUVMask(0xff00ff00ff00ff00i64);
					Iu8vec8 mZero(_mm_setzero_si64());

					Iu8vec8 mSrc_H = _m_from_BYTEs(pSrc + nX);
					Iu8vec8 mMiniSrc_H = _m_from_BYTEs(pMiniSrc);
					Iu16vec4 mEdge_H = _m_from_BYTEs(pEdgeBuffer + nX);

					Iu8vec8 mBlendColor_H = (((Iu16vec4(mSrc_H & mUVMask) >> 8) + (Iu16vec4(mMiniSrc_H & mUVMask) >> 8)) >> 1) << 8;
					_m_from_BYTEs(pDest + nX) = select_eq(Iu8vec8(mEdge_H << 8), mZero, mSrc_H, mBlendColor_H);

					Iu8vec8 mSrc_L = _m_from_BYTEs(pSrc + nX + 8);
					Iu8vec8 mMiniSrc_L = _m_from_BYTEs(pMiniSrc + 8);
					Iu16vec4 mEdge_L = _m_from_BYTEs(pEdgeBuffer + nX + 8);

					Iu8vec8 mBlendColor_L = (((Iu16vec4(mSrc_L & mUVMask) >> 8) + (Iu16vec4(mMiniSrc_L & mUVMask) >> 8)) >> 1) << 8;
					_m_from_BYTEs(pDest + nX + 8) = select_eq(Iu8vec8(mEdge_L << 8), mZero, mSrc_L, mBlendColor_L);
				}
			}
		}

		pSrcTT += nSrcPitch;
		pSrcT += nSrcPitch;
		pSrc += nSrcPitch;
		pSrcB += nSrcPitch;
		pSrcBB += nSrcPitch;
		pSrcPTT += nSrcPitch;
		pSrcPT += nSrcPitch;
		pSrcP += nSrcPitch;
		pSrcPB += nSrcPitch;
		pSrcPBB += nSrcPitch;
		pSrcFTT += nSrcPitch;
		pSrcFT += nSrcPitch;
		pSrcF += nSrcPitch;
		pSrcFB += nSrcPitch;
		pSrcFBB += nSrcPitch;

		pDest += nDestPitch;
	}

	delete[] pEdgeBuffer;

	empty();

	return destFrame;
}

// Px̍
void DeCross::Diff_YUY2(const BYTE* pDiff0, const BYTE* pDiff1, const int nPos, int& nMiniDiff, const BYTE* &pMiniSrc)
{
	I8vec8 mYMask(0x00ff00ff00ff00ffi64);

	Is16vec4 mDiff0_H = _m_from_BYTEs(pDiff0 + nPos * 2) & mYMask;
	Is16vec4 mDiff1_H = _m_from_BYTEs(pDiff1) & mYMask;
	Is16vec4 mDiff0_L = _m_from_BYTEs(pDiff0 + 8 + nPos * 2) & mYMask;
	Is16vec4 mDiff1_L = _m_from_BYTEs(pDiff1 + 8) & mYMask;

	Is16vec4 mDiff = abs(mDiff0_H - mDiff1_H) + abs(mDiff0_L - mDiff1_L);
	int nDiff = mDiff[0] + mDiff[1] + mDiff[2] + mDiff[3];
	if (nDiff < nMiniDiff)
	{
		pMiniSrc = pDiff0 + nPos * 2;
		nMiniDiff = nDiff;
	}
}

// GbWo
void DeCross::EdgeCheck_YUY2(const BYTE* pSrc, BYTE* pEdgeBuffer, const int nRowSize)
{
	BYTE YThreshold[8];
	YThreshold[0] = YThreshold[2] = YThreshold[4] = YThreshold[6] = (BYTE)m_nYThreshold;
	YThreshold[1] = YThreshold[3] = YThreshold[5] = YThreshold[7] = 0;

	Is16vec4 mYThreshold(_m_from_BYTEs(YThreshold));
	I8vec8 mYMask(0x00ff00ff00ff00ffi64);

	int nMarginMax = m_nMargin << 2;

	int nX;
	for (nX = 16; nX < nRowSize - 16; nX += 8)
	{
		Is16vec4 mLeft = _m_from_BYTEs(pSrc + nX - 2) & mYMask;
		Is16vec4 mCenter = _m_from_BYTEs(pSrc + nX) & mYMask;
		Is16vec4 mRight = _m_from_BYTEs(pSrc + nX + 2) & mYMask;

		Iu32vec2 mEdge =
			(mYThreshold < abs(mLeft - mRight)) &
			((mLeft < mCenter & mCenter < mRight) |
			(mLeft > mCenter & mCenter > mRight));
		mEdge |= (mEdge << 16) | (mEdge >> 16);

		int nMargin;
		for (nMargin = -nMarginMax; nMargin <= nMarginMax; nMargin += 4)
		{
			_m_from_BYTEs(pEdgeBuffer + nX + nMargin) = _m_from_BYTEs(pEdgeBuffer + nX + nMargin) | mEdge;
		}
	}
}

PVideoFrame DeCross::DeCross_YV12(int n, IScriptEnvironment *env)
{
	PVideoFrame srcFrame = child->GetFrame(n, env);
	PVideoFrame destFrame = srcFrame;
	env->MakeWritable(&destFrame);

	const int nHeightU = srcFrame->GetHeight(PLANAR_U);
	const int nRowSizeU = srcFrame->GetRowSize(PLANAR_U);
	const int nSrcPitch = srcFrame->GetPitch();
	const int nSrcPitch2 = nSrcPitch * 2;
	const int nSrcPitchU = srcFrame->GetPitch(PLANAR_U);
	const int nDestPitchU = destFrame->GetPitch(PLANAR_U);

	const BYTE* pSrc = srcFrame->GetReadPtr() + nSrcPitch2;
	const BYTE* pSrcP = child->GetFrame(n - 1, env)->GetReadPtr() + nSrcPitch2;
	const BYTE* pSrcF = child->GetFrame(n + 1, env)->GetReadPtr() + nSrcPitch2;

	const BYTE* pSrcTT = pSrc - nSrcPitch2;
	const BYTE* pSrcBB = pSrc + nSrcPitch2;
	const BYTE* pSrcPTT = pSrcP - nSrcPitch2;
	const BYTE* pSrcPBB = pSrcP + nSrcPitch2;
	const BYTE* pSrcFTT = pSrcF - nSrcPitch2;
	const BYTE* pSrcFBB = pSrcF + nSrcPitch2;

	const BYTE* pSrcT = pSrc - nSrcPitch;
	const BYTE* pSrcB = pSrc + nSrcPitch;
	const BYTE* pSrcPT = pSrcP - nSrcPitch;
	const BYTE* pSrcPB = pSrcP + nSrcPitch;
	const BYTE* pSrcFT = pSrcF - nSrcPitch;
	const BYTE* pSrcFB = pSrcF + nSrcPitch;

	const BYTE* pSrcU = srcFrame->GetReadPtr(PLANAR_U) + nSrcPitchU;
	const BYTE* pSrcUP = child->GetFrame(n - 1, env)->GetReadPtr(PLANAR_U) + nSrcPitchU;
	const BYTE* pSrcUF = child->GetFrame(n + 1, env)->GetReadPtr(PLANAR_U) + nSrcPitchU;
	const BYTE* pSrcV = srcFrame->GetReadPtr(PLANAR_V) + nSrcPitchU;
	const BYTE* pSrcVP = child->GetFrame(n - 1, env)->GetReadPtr(PLANAR_V) + nSrcPitchU;
	const BYTE* pSrcVF = child->GetFrame(n + 1, env)->GetReadPtr(PLANAR_V) + nSrcPitchU;

	const BYTE* pSrcUTT = pSrcU - nSrcPitchU;
	const BYTE* pSrcUBB = pSrcU + nSrcPitchU;
	const BYTE* pSrcUPTT = pSrcUP - nSrcPitchU;
	const BYTE* pSrcUPBB = pSrcUP + nSrcPitchU;
	const BYTE* pSrcUFTT = pSrcUF - nSrcPitchU;
	const BYTE* pSrcUFBB = pSrcUF + nSrcPitchU;
	const BYTE* pSrcVTT = pSrcV - nSrcPitchU;
	const BYTE* pSrcVBB = pSrcV + nSrcPitchU;
	const BYTE* pSrcVPTT = pSrcVP - nSrcPitchU;
	const BYTE* pSrcVPBB = pSrcVP + nSrcPitchU;
	const BYTE* pSrcVFTT = pSrcVF - nSrcPitchU;
	const BYTE* pSrcVFBB = pSrcVF + nSrcPitchU;

	const BYTE* pSrcUMini;
	const BYTE* pSrcVMini;

	BYTE* pDestU = destFrame->GetWritePtr(PLANAR_U) + nDestPitchU;
	BYTE* pDestV = destFrame->GetWritePtr(PLANAR_V) + nDestPitchU;

	BYTE* pEdgeBuffer = new BYTE[nRowSizeU];

	int nX, nY;
	for (nY = nHeightU - 1; nY > 1; nY--)
	{
		::ZeroMemory(pEdgeBuffer, nRowSizeU);

		EdgeCheck_YV12(pSrc, pEdgeBuffer, nRowSizeU);

		if (m_bDebug)
		{
			for (nX = 4; nX < nRowSizeU - 4; nX++)
			{
				if (pEdgeBuffer[nX] != 0)
				{
					pDestU[nX] = 128;
					pDestV[nX] = 255;
				}
			}
		}
		else
		{
			int nX2 = 0;
			for (nX = 4; nX < nRowSizeU - 4; nX += 4)
			{
				nX2 += 4 * 2;
				if(*(DWORD*)&pEdgeBuffer[nX] != 0)
				{
					int nMiniDiff = m_nNoiseThreshold;
					pSrcUMini = pSrcU;
					pSrcVMini = pSrcV;

					if (nY % 2 == 1)
					{
						if (Diff_YV12(pSrcPTT + nX2, pSrcT + nX2, -6, nMiniDiff)) { pSrcUMini = pSrcUPTT - 3; pSrcVMini = pSrcVPTT - 3; }
						if (Diff_YV12(pSrcPTT + nX2, pSrcT + nX2, -2, nMiniDiff)) { pSrcUMini = pSrcUPTT - 1; pSrcVMini = pSrcVPTT - 1; }
						if (Diff_YV12(pSrcPT  + nX2, pSrcT + nX2, -4, nMiniDiff)) { pSrcUMini = pSrcUP   - 2; pSrcVMini = pSrcVP   - 2; }
						if (Diff_YV12(pSrcPBB + nX2, pSrcT + nX2, -6, nMiniDiff)) { pSrcUMini = pSrcUPBB - 3; pSrcVMini = pSrcVPBB - 3; }
						if (Diff_YV12(pSrcPBB + nX2, pSrcT + nX2, -2, nMiniDiff)) { pSrcUMini = pSrcUPBB - 1; pSrcVMini = pSrcVPBB - 1; }

						if (Diff_YV12(pSrcTT + nX2, pSrcT + nX2, -6, nMiniDiff)) { pSrcUMini = pSrcUTT - 3; pSrcVMini = pSrcVTT - 3; }
						if (Diff_YV12(pSrcTT + nX2, pSrcT + nX2, -2, nMiniDiff)) { pSrcUMini = pSrcUTT - 1; pSrcVMini = pSrcVTT - 1; }
						if (Diff_YV12(pSrcT  + nX2, pSrcT + nX2, -4, nMiniDiff)) { pSrcUMini = pSrcU   - 2; pSrcVMini = pSrcV   - 2; }
						if (Diff_YV12(pSrcBB + nX2, pSrcT + nX2, -6, nMiniDiff)) { pSrcUMini = pSrcUBB - 3; pSrcVMini = pSrcVBB - 3; }
						if (Diff_YV12(pSrcBB + nX2, pSrcT + nX2, -2, nMiniDiff)) { pSrcUMini = pSrcUBB - 1; pSrcVMini = pSrcVBB - 1; }

						if (Diff_YV12(pSrcFTT + nX2, pSrcT + nX2, -6, nMiniDiff)) { pSrcUMini = pSrcUFTT - 3; pSrcVMini = pSrcVFTT - 3; }
						if (Diff_YV12(pSrcFTT + nX2, pSrcT + nX2, -2, nMiniDiff)) { pSrcUMini = pSrcUFTT - 1; pSrcVMini = pSrcVFTT - 1; }
						if (Diff_YV12(pSrcFT  + nX2, pSrcT + nX2, -4, nMiniDiff)) { pSrcUMini = pSrcUF   - 2; pSrcVMini = pSrcVF   - 2; }
						if (Diff_YV12(pSrcFBB + nX2, pSrcT + nX2, -6, nMiniDiff)) { pSrcUMini = pSrcUFBB - 3; pSrcVMini = pSrcVFBB - 3; }
						if (Diff_YV12(pSrcFBB + nX2, pSrcT + nX2, -2, nMiniDiff)) { pSrcUMini = pSrcUFBB - 1; pSrcVMini = pSrcVFBB - 1; }

						if (Diff_YV12(pSrcPT + nX2, pSrcT + nX2, -0, nMiniDiff)) { pSrcUMini = pSrcUP - 0; pSrcVMini = pSrcVP - 0; }
						if (Diff_YV12(pSrcFT + nX2, pSrcT + nX2, -0, nMiniDiff)) { pSrcUMini = pSrcUF - 0; pSrcVMini = pSrcVF - 0; }
						if (Diff_YV12(pSrcPB + nX2, pSrcB + nX2, -0, nMiniDiff)) { pSrcUMini = pSrcUP - 0; pSrcVMini = pSrcVP - 0; }
						if (Diff_YV12(pSrcFB + nX2, pSrcB + nX2, -0, nMiniDiff)) { pSrcUMini = pSrcUF - 0; pSrcVMini = pSrcVF - 0; }

						if (Diff_YV12(pSrcPTT + nX2, pSrcT + nX2, +6, nMiniDiff)) { pSrcUMini = pSrcUPTT + 3; pSrcVMini = pSrcVPTT + 3; }
						if (Diff_YV12(pSrcPTT + nX2, pSrcT + nX2, +2, nMiniDiff)) { pSrcUMini = pSrcUPTT + 1; pSrcVMini = pSrcVPTT + 1; }
						if (Diff_YV12(pSrcPT  + nX2, pSrcT + nX2, +4, nMiniDiff)) { pSrcUMini = pSrcUP   + 2; pSrcVMini = pSrcVP   + 2; }
						if (Diff_YV12(pSrcPBB + nX2, pSrcT + nX2, +6, nMiniDiff)) { pSrcUMini = pSrcUPBB + 3; pSrcVMini = pSrcVPBB + 3; }
						if (Diff_YV12(pSrcPBB + nX2, pSrcT + nX2, +2, nMiniDiff)) { pSrcUMini = pSrcUPBB + 1; pSrcVMini = pSrcVPBB + 1; }

						if (Diff_YV12(pSrcTT + nX2, pSrcT + nX2, +6, nMiniDiff)) { pSrcUMini = pSrcUTT + 3; pSrcVMini = pSrcVTT + 3; }
						if (Diff_YV12(pSrcTT + nX2, pSrcT + nX2, +2, nMiniDiff)) { pSrcUMini = pSrcUTT + 1; pSrcVMini = pSrcVTT + 1; }
						if (Diff_YV12(pSrcT  + nX2, pSrcT + nX2, +4, nMiniDiff)) { pSrcUMini = pSrcU   + 2; pSrcVMini = pSrcV   + 2; }
						if (Diff_YV12(pSrcBB + nX2, pSrcT + nX2, +6, nMiniDiff)) { pSrcUMini = pSrcUBB + 3; pSrcVMini = pSrcVBB + 3; }
						if (Diff_YV12(pSrcBB + nX2, pSrcT + nX2, +2, nMiniDiff)) { pSrcUMini = pSrcUBB + 1; pSrcVMini = pSrcVBB + 1; }

						if (Diff_YV12(pSrcFTT + nX2, pSrcT + nX2, +6, nMiniDiff)) { pSrcUMini = pSrcUFTT + 3; pSrcVMini = pSrcVFTT + 3; }
						if (Diff_YV12(pSrcFTT + nX2, pSrcT + nX2, +2, nMiniDiff)) { pSrcUMini = pSrcUFTT + 1; pSrcVMini = pSrcVFTT + 1; }
						if (Diff_YV12(pSrcFT  + nX2, pSrcT + nX2, +4, nMiniDiff)) { pSrcUMini = pSrcUF   + 2; pSrcVMini = pSrcVF   + 2; }
						if (Diff_YV12(pSrcFBB + nX2, pSrcT + nX2, +6, nMiniDiff)) { pSrcUMini = pSrcUFBB + 3; pSrcVMini = pSrcVFBB + 3; }
						if (Diff_YV12(pSrcFBB + nX2, pSrcT + nX2, +2, nMiniDiff)) { pSrcUMini = pSrcUFBB + 1; pSrcVMini = pSrcVFBB + 1; }
					}
					else
					{
						if (Diff_YV12(pSrcPT + nX2, pSrc + nX2, -6, nMiniDiff)) { pSrcUMini = pSrcUPTT - 3; pSrcVMini = pSrcVPTT - 3; }
						if (Diff_YV12(pSrcPT + nX2, pSrc + nX2, -2, nMiniDiff)) { pSrcUMini = pSrcUPTT - 1; pSrcVMini = pSrcVPTT - 1; }
						if (Diff_YV12(pSrcP  + nX2, pSrc + nX2, -4, nMiniDiff)) { pSrcUMini = pSrcUP   - 2; pSrcVMini = pSrcVP   - 2; }
						if (Diff_YV12(pSrcPB + nX2, pSrc + nX2, -6, nMiniDiff)) { pSrcUMini = pSrcUPBB - 3; pSrcVMini = pSrcVPBB - 3; }
						if (Diff_YV12(pSrcPB + nX2, pSrc + nX2, -2, nMiniDiff)) { pSrcUMini = pSrcUPBB - 1; pSrcVMini = pSrcVPBB - 1; }

						if (Diff_YV12(pSrcT + nX2, pSrc + nX2, -6, nMiniDiff)) { pSrcUMini = pSrcUTT - 3; pSrcVMini = pSrcVTT - 3; }
						if (Diff_YV12(pSrcT + nX2, pSrc + nX2, -2, nMiniDiff)) { pSrcUMini = pSrcUTT - 1; pSrcVMini = pSrcVTT - 1; }
						if (Diff_YV12(pSrc  + nX2, pSrc + nX2, -4, nMiniDiff)) { pSrcUMini = pSrcU   - 2; pSrcVMini = pSrcV   - 2; }
						if (Diff_YV12(pSrcB + nX2, pSrc + nX2, -6, nMiniDiff)) { pSrcUMini = pSrcUBB - 3; pSrcVMini = pSrcVBB - 3; }
						if (Diff_YV12(pSrcB + nX2, pSrc + nX2, -2, nMiniDiff)) { pSrcUMini = pSrcUBB - 1; pSrcVMini = pSrcVBB - 1; }

						if (Diff_YV12(pSrcFT + nX2, pSrc + nX2, -6, nMiniDiff)) { pSrcUMini = pSrcUFTT - 3; pSrcVMini = pSrcVFTT - 3; }
						if (Diff_YV12(pSrcFT + nX2, pSrc + nX2, -2, nMiniDiff)) { pSrcUMini = pSrcUFTT - 1; pSrcVMini = pSrcVFTT - 1; }
						if (Diff_YV12(pSrcF  + nX2, pSrc + nX2, -4, nMiniDiff)) { pSrcUMini = pSrcUF   - 2; pSrcVMini = pSrcVF   - 2; }
						if (Diff_YV12(pSrcFB + nX2, pSrc + nX2, -6, nMiniDiff)) { pSrcUMini = pSrcUFBB - 3; pSrcVMini = pSrcVFBB - 3; }
						if (Diff_YV12(pSrcFB + nX2, pSrc + nX2, -2, nMiniDiff)) { pSrcUMini = pSrcUFBB - 1; pSrcVMini = pSrcVFBB - 1; }

						if (Diff_YV12(pSrcP + nX2, pSrc + nX2, -0, nMiniDiff)) { pSrcUMini = pSrcUP - 0; pSrcVMini = pSrcVP - 0; }
						if (Diff_YV12(pSrcF + nX2, pSrc + nX2, -0, nMiniDiff)) { pSrcUMini = pSrcUF - 0; pSrcVMini = pSrcVF - 0; }

						if (Diff_YV12(pSrcPT + nX2, pSrc + nX2, +6, nMiniDiff)) { pSrcUMini = pSrcUPTT + 3; pSrcVMini = pSrcVPTT + 3; }
						if (Diff_YV12(pSrcPT + nX2, pSrc + nX2, +2, nMiniDiff)) { pSrcUMini = pSrcUPTT + 1; pSrcVMini = pSrcVPTT + 1; }
						if (Diff_YV12(pSrcP  + nX2, pSrc + nX2, +4, nMiniDiff)) { pSrcUMini = pSrcUP   + 2; pSrcVMini = pSrcVP   + 2; }
						if (Diff_YV12(pSrcPB + nX2, pSrc + nX2, +6, nMiniDiff)) { pSrcUMini = pSrcUPBB + 3; pSrcVMini = pSrcVPBB + 3; }
						if (Diff_YV12(pSrcPB + nX2, pSrc + nX2, +2, nMiniDiff)) { pSrcUMini = pSrcUPBB + 1; pSrcVMini = pSrcVPBB + 1; }

						if (Diff_YV12(pSrcT + nX2, pSrc + nX2, +6, nMiniDiff)) { pSrcUMini = pSrcUTT + 3; pSrcVMini = pSrcVTT + 3; }
						if (Diff_YV12(pSrcT + nX2, pSrc + nX2, +2, nMiniDiff)) { pSrcUMini = pSrcUTT + 1; pSrcVMini = pSrcVTT + 1; }
						if (Diff_YV12(pSrc  + nX2, pSrc + nX2, +4, nMiniDiff)) { pSrcUMini = pSrcU   + 2; pSrcVMini = pSrcV   + 2; }
						if (Diff_YV12(pSrcB + nX2, pSrc + nX2, +6, nMiniDiff)) { pSrcUMini = pSrcUBB + 3; pSrcVMini = pSrcVBB + 3; }
						if (Diff_YV12(pSrcB + nX2, pSrc + nX2, +2, nMiniDiff)) { pSrcUMini = pSrcUBB + 1; pSrcVMini = pSrcVBB + 1; }

						if (Diff_YV12(pSrcFT + nX2, pSrc + nX2, +6, nMiniDiff)) { pSrcUMini = pSrcUFTT + 3; pSrcVMini = pSrcVFTT + 3; }
						if (Diff_YV12(pSrcFT + nX2, pSrc + nX2, +2, nMiniDiff)) { pSrcUMini = pSrcUFTT + 1; pSrcVMini = pSrcVFTT + 1; }
						if (Diff_YV12(pSrcF  + nX2, pSrc + nX2, +4, nMiniDiff)) { pSrcUMini = pSrcUF   + 2; pSrcVMini = pSrcVF   + 2; }
						if (Diff_YV12(pSrcFB + nX2, pSrc + nX2, +6, nMiniDiff)) { pSrcUMini = pSrcUFBB + 3; pSrcVMini = pSrcVFBB + 3; }
						if (Diff_YV12(pSrcFB + nX2, pSrc + nX2, +2, nMiniDiff)) { pSrcUMini = pSrcUFBB + 1; pSrcVMini = pSrcVFBB + 1; }
					}

					Iu8vec8 mZero(_mm_setzero_si64());

					Iu8vec8 mSrcU = _m_from_BYTEs(pSrcU + nX);
					Iu8vec8 mSrcV = _m_from_BYTEs(pSrcV + nX);
					Iu16vec4 mSrcULow = unpack_low(_m_from_BYTEs(pSrcU + nX), mZero);
					Iu16vec4 mSrcVLow = unpack_low(_m_from_BYTEs(pSrcV + nX), mZero);
					Iu16vec4 mSrcUMiniLow = unpack_low(_m_from_BYTEs(pSrcUMini + nX), mZero);
					Iu16vec4 mSrcVMiniLow = unpack_low(_m_from_BYTEs(pSrcVMini + nX), mZero);
					I8vec8 mLowMask(0x00000000ffffffffi64);
					Iu8vec8 mEdgeLow = _m_from_BYTEs(pEdgeBuffer + nX) & mLowMask;

					Iu8vec8 mBlendColorU = packu_sat(Is16vec4((mSrcULow + mSrcUMiniLow) >> 1), Is16vec4(mZero));
					_m_from_BYTEs(pDestU + nX) = select_eq(mEdgeLow, mZero, mSrcU, mBlendColorU);

					Iu8vec8 mBlendColorV = packu_sat(Is16vec4((mSrcVLow + mSrcVMiniLow) >> 1), Is16vec4(mZero));
					_m_from_BYTEs(pDestV + nX) = select_eq(mEdgeLow, mZero, mSrcV, mBlendColorV);
				}
			}
		}
		pSrc += nSrcPitch2;
		pSrcP += nSrcPitch2;
		pSrcF += nSrcPitch2;

		pSrcTT += nSrcPitch2;
		pSrcBB += nSrcPitch2;
		pSrcPTT += nSrcPitch2;
		pSrcPBB += nSrcPitch2;
		pSrcFTT += nSrcPitch2;
		pSrcFBB += nSrcPitch2;

		pSrcT += nSrcPitch2;
		pSrcB += nSrcPitch2;
		pSrcPT += nSrcPitch2;
		pSrcPB += nSrcPitch2;
		pSrcFT += nSrcPitch2;
		pSrcFB += nSrcPitch2;

		pSrcU += nSrcPitchU;
		pSrcUP += nSrcPitchU;
		pSrcUF += nSrcPitchU;
		pSrcV += nSrcPitchU;
		pSrcVP += nSrcPitchU;
		pSrcVF += nSrcPitchU;

		pSrcUTT += nSrcPitchU;
		pSrcUBB += nSrcPitchU;
		pSrcUPTT += nSrcPitchU;
		pSrcUPBB += nSrcPitchU;
		pSrcUFTT += nSrcPitchU;
		pSrcUFBB += nSrcPitchU;
		pSrcVTT += nSrcPitchU;
		pSrcVBB += nSrcPitchU;
		pSrcVPTT += nSrcPitchU;
		pSrcVPBB += nSrcPitchU;
		pSrcVFTT += nSrcPitchU;
		pSrcVFBB += nSrcPitchU;

		pDestU += nDestPitchU;
		pDestV += nDestPitchU;
	}

	delete[] pEdgeBuffer;

	empty();

	return destFrame;
}

// Px̍
bool DeCross::Diff_YV12(const BYTE* pDiff0, const BYTE* pDiff1, const int nPos, int& nMiniDiff)
{
	I8vec8 mZero(_mm_setzero_si64());

	I8vec8 mDiff0 = _m_from_BYTEs(pDiff0 + nPos);
	I8vec8 mDiff1 = _m_from_BYTEs(pDiff1);
	Is16vec4 mDiff0_H = unpack_high(mDiff0, mZero);
	Is16vec4 mDiff1_H = unpack_high(mDiff1, mZero);
	Is16vec4 mDiff0_L = unpack_low(mDiff0, mZero);
	Is16vec4 mDiff1_L = unpack_low(mDiff1, mZero);

	Is16vec4 mDiff = abs(mDiff0_H - mDiff1_H) + abs(mDiff0_L - mDiff1_L);
	int nDiff = mDiff[0] + mDiff[1] + mDiff[2] + mDiff[3];
	if (nDiff < nMiniDiff)
	{
		nMiniDiff = nDiff;
		return true;
	}
	return false;
}

// GbWo
void DeCross::EdgeCheck_YV12(const BYTE* pSrc, BYTE* pEdgeBuffer, const int nRowSizeU)
{
	BYTE YThreshold[8];
	YThreshold[0] = YThreshold[2] = YThreshold[4] = YThreshold[6] = (BYTE)m_nYThreshold;
	YThreshold[1] = YThreshold[3] = YThreshold[5] = YThreshold[7] = 0;

	Is16vec4 mYThreshold(_m_from_BYTEs(YThreshold));
	I8vec8 mYMask(0x00ff00ff00ff00ffi64);
	Is16vec4 mZero(_mm_setzero_si64());

	int nX;
	for (nX = 4; nX < nRowSizeU - 4; nX += 4)
	{
		Is16vec4 mLeft = _m_from_BYTEs(pSrc + nX * 2 - 1);
		Is16vec4 mCenter = _m_from_BYTEs(pSrc + nX * 2);
		Is16vec4 mRight = _m_from_BYTEs(pSrc + nX * 2 + 1);

		Is16vec4 mLeft1 = mLeft & mYMask;
		Is16vec4 mCenter1 = mCenter & mYMask;
		Is16vec4 mRight1 = mRight & mYMask;

		Is16vec4 mEdge1 =
			(mYThreshold < abs(mLeft1 - mRight1)) &
			((mLeft1 < mCenter1 & mCenter1 < mRight1) |
			(mLeft1 > mCenter1 & mCenter1 > mRight1));

		Is16vec4 mEdge = pack_sat(mEdge1, mZero);

		Is16vec4 mLeft2 = (mLeft >> 8) & mYMask;
		Is16vec4 mCenter2 = (mCenter >> 8) & mYMask;
		Is16vec4 mRight2 = (mRight >> 8) & mYMask;

		Is16vec4 mEdge2 =
			mYThreshold < abs(mLeft2 - mRight2) &
			((mLeft2 < mCenter2 & mCenter2 < mRight2) |
			(mLeft2 > mCenter2 & mCenter2 > mRight2));

		mEdge |= pack_sat(mEdge2, mZero);

		int nMargin;
		for (nMargin = -m_nMargin; nMargin <= m_nMargin; nMargin++)
		{
			_m_from_BYTEs(pEdgeBuffer + nX + nMargin) = _m_from_BYTEs(pEdgeBuffer + nX + nMargin) | mEdge;
		}
	}
}
