#ifndef IVEC_WRAPPER_H_INCLUDED
#define IVEC_WRAPPER_H_INCLUDED

#pragma warning (push, 1)
#pragma warning (disable : 4799)
#include <ivec.h>
#pragma warning (pop)

#define _m_from_BYTEs(b) (*(__m64*)(b))

inline Is8vec8 sat_adds(const I8vec8 &a, const I8vec8 &b)		{ return _m_paddsb(a,b); }
inline Is8vec8 sat_subs(const I8vec8 &a, const I8vec8 &b)		{ return _m_psubsb(a,b); }
inline Iu8vec8 sat_addu(const I8vec8 &a, const I8vec8 &b)		{ return _m_paddusb(a,b); }
inline Iu8vec8 sat_subu(const I8vec8 &a, const I8vec8 &b)		{ return _m_psubusb(a,b); }
//inline Is8vec8 sat_adds(const I8vec8 &a, const I8vec8 &b)		{ return sat_add((Is8vec8)a,(Is8vec8)b); }
//inline Is8vec8 sat_subs(const I8vec8 &a, const I8vec8 &b)		{ return sat_sub((Is8vec8)a,(Is8vec8)b); }
//inline Iu8vec8 sat_addu(const I8vec8 &a, const I8vec8 &b)		{ return sat_add((Is8vec8)a,(Is8vec8)b); }
//inline Iu8vec8 sat_subu(const I8vec8 &a, const I8vec8 &b)		{ return sat_sub((Is8vec8)a,(Is8vec8)b); }

inline __m64 _mm_setallones_si64() { I8vec8 m; return _m_pcmpeqd(m, m); }

inline Is8vec8 abs(const Is8vec8 &a)
{
	Is8vec8 r0 = a;
	Is8vec8 r1 = a;
	Is8vec8 zero = _mm_setzero_si64();
	r1 = cmplt(r1, zero);
	r0 ^= r1;
	r0 -= r1;
	return r0;
}

inline Is16vec4 abs(const Is16vec4 &a)
{
	Is16vec4 r0 = a;
	Is16vec4 r1 = a;
	Is16vec4 zero = _mm_setzero_si64();
	r1 = cmplt(r1, zero);
	r0 ^= r1;
	r0 -= r1;
	return r0;
}

inline Is8vec8 operator==(const Is8vec8 &a, const Is8vec8 &b)		{ return _m_pcmpeqb(a,b); }
inline Is8vec8 operator!=(const Is8vec8 &a, const Is8vec8 &b)		{ return _m_pandn(_m_pcmpeqb(a,b), M64(0xffffffffffffffffi64)); }
inline Is8vec8 operator>(const Is8vec8 &a, const Is8vec8 &b)		{ return _m_pcmpgtb(a,b); }
inline Is8vec8 operator<(const Is8vec8 &a, const Is8vec8 &b)		{ return _m_pcmpgtb(b,a); }
inline Is8vec8 operator<=(const Is8vec8 &a, const Is8vec8 &b)		{ return _m_pandn(_m_pcmpgtb(a,b), M64(0xffffffffffffffffi64)); }
inline Is8vec8 operator>=(const Is8vec8 &a, const Is8vec8 &b)		{ return _m_pandn(_m_pcmpgtb(b,a), M64(0xffffffffffffffffi64)); }

inline Is16vec4 operator==(const Is16vec4 &a, const Is16vec4 &b)    { return _m_pcmpeqw(a,b); }
inline Is16vec4 operator!=(const Is16vec4 &a, const Is16vec4 &b)    { return _m_pandn(_m_pcmpeqw(a,b), M64(0xffffffffffffffffi64)); }
inline Is16vec4 operator>(const Is16vec4 &a, const Is16vec4 &b)		{ return _m_pcmpgtw(a,b); }
inline Is16vec4 operator<(const Is16vec4 &a, const Is16vec4 &b)		{ return _m_pcmpgtw(b,a); }
inline Is16vec4 operator<=(const Is16vec4 &a, const Is16vec4 &b)	{ return _m_pandn(_m_pcmpgtw(a,b), M64(0xffffffffffffffffi64)); }
inline Is16vec4 operator>=(const Is16vec4 &a, const Is16vec4 &b)	{ return _m_pandn(_m_pcmpgtw(b,a), M64(0xffffffffffffffffi64)); }

#endif // IVEC_WRAPPER_H_INCLUDED
